# -*- encoding: utf-8
"""
A miscellaneous collection of functions and helpers for use in the CLI.
Named junkdrawer.py because I haven't organised it properly.

https://twitter.com/betsythemuffin/status/1003313844108824584

"""

import datetime as dt
import os
import re
import sys

from .superprocess import call, check_call, check_output


ROOT = check_output('git', 'rev-parse', '--show-toplevel')

RELEASE_FILE = os.path.join(ROOT, 'RELEASE.md')

MAJOR = 'major'
MINOR = 'minor'
PATCH = 'patch'

VALID_RELEASE_TYPES = (MAJOR, MINOR, PATCH)

RELEASE_TYPE = re.compile(r"^RELEASE_TYPE: +(%s|%s|%s)" % VALID_RELEASE_TYPES)

CHANGELOG_HEADER = re.compile(r"^## v\d+\.\d+\.\d+ - \d\d\d\d-\d\d-\d\d$")
CHANGELOG_FILE = os.path.join(ROOT, 'CHANGELOG.md')

BUILD_SBT = os.path.join(ROOT, 'build.sbt')


def git(*args):
    check_call('git', *args)


def hash_for_name(name):
    return check_output('git', 'rev-parse', '--name')


def is_ancestor(a, b):
    check = call('git', 'merge-base', '--is-ancestor', a, b)
    assert 0 <= check <= 1
    return check == 0


def get_changed_paths(*args):
    """
    Returns a set of changed paths in a given commit range.

    :param commit_range: Arguments to pass to ``git diff``.
    """
    diff_output = check_output('git', 'diff', '--name-only', *args)

    return set([
        line.strip()
        for line in diff_output.splitlines()
    ])


def run_formatting(run_all=False):
    """
    Run all the formatting scripts.
    """
    if not run_all:
        changed_paths = get_changed_paths('HEAD', 'master')

    if run_all or any(f.endswith(('.sbt', '.scala')) for f in changed_paths):
        check_call('sbt', 'scalafmt')

    if run_all or any(f.endswith('.tf') for f in changed_paths):
        check_call('terraform', 'fmt')


def tags():
    """
    Returns a list of all tags in the repo.
    """
    git('fetch', '--tags')
    result = check_output('git', 'tag')
    all_tags = result.split('\n')

    assert len(set(all_tags)) == len(all_tags)

    return set(all_tags)


def latest_version():
    """
    Returns the latest version, as specified by the Git tags.
    """
    versions = []

    all_tags = tags()

    for t in all_tags:
        assert t == t.strip()
        parts = t.split('.')
        assert len(parts) == 3, t
        parts[0] = parts[0].lstrip('v')
        v = tuple(map(int, parts))

        versions.append((v, t))

    _, latest = max(versions)

    assert latest in all_tags
    return latest


def modified_files():
    """
    Returns a list of all files which have been modified between now
    and the latest release.
    """
    files = set()
    for command in [
        ('git', 'diff', '--name-only', '--diff-filter=d',
            latest_version(), 'HEAD'),
        ('git', 'diff', '--name-only')
    ]:
        diff_output = check_output(*command)
        for l in diff_output.split('\n'):
            filepath = l.strip()
            if filepath:
                assert os.path.exists(filepath)
                files.add(filepath)
    return files


def has_source_changes():
    """
    Returns True if there are source changes since the previous release,
    False if not.
    """
    changed_files = [
        f for f in modified_files() if f.strip().endswith(('.sbt', '.scala'))
    ]
    return len(changed_files) != 0


def has_release():
    """
    Returns True if there is a release file, False if not.
    """
    return os.path.exists(RELEASE_FILE)


def parse_release_file():
    """
    Parses the release file, returning a tuple (release_type, release_contents)
    """
    with open(RELEASE_FILE) as i:
        release_contents = i.read()

    release_lines = release_contents.split('\n')

    m = RELEASE_TYPE.match(release_lines[0])
    if m is not None:
        release_type = m.group(1)
        if release_type not in VALID_RELEASE_TYPES:
            print('Unrecognised release type %r' % (release_type,))
            sys.exit(1)
        del release_lines[0]
        release_contents = '\n'.join(release_lines).strip()
    else:
        print(
            'RELEASE.md does not start by specifying release type. The first '
            'line of the file should be RELEASE_TYPE: followed by one of '
            'major, minor, or patch, to specify the type of release that '
            'this is (i.e. which version number to increment). Instead the '
            'first line was %r' % (release_lines[0],)
        )
        sys.exit(1)

    return release_type, release_contents


def changelog():
    with open(CHANGELOG_FILE) as i:
        return i.read()


def new_version(release_type):
    version = latest_version()
    version_info = [int(i) for i in version.lstrip('v').split('.')]

    new_version = list(version_info)
    bump = VALID_RELEASE_TYPES.index(release_type)
    new_version[bump] += 1
    for i in range(bump + 1, len(new_version)):
        new_version[i] = 0
    new_version = tuple(new_version)
    return 'v' + '.'.join(map(str, new_version))


def update_changelog_and_scala_version():
    contents = changelog()
    assert '\r' not in contents
    lines = contents.split('\n')
    assert contents == '\n'.join(lines)
    for i, l in enumerate(lines):
        if CHANGELOG_HEADER.match(l):
            beginning = '\n'.join(lines[:i])
            rest = '\n'.join(lines[i:])
            assert '\n'.join((beginning, rest)) == contents
            break

    release_type, release_contents = parse_release_file()

    new_version_string = new_version(release_type)

    print('New version: %s' % new_version_string)

    now = dt.datetime.utcnow()

    date = max([
        d.strftime('%Y-%m-%d') for d in (now, now + dt.timedelta(hours=1))
    ])

    heading_for_new_version = '## ' + ' - '.join((new_version_string, date))

    new_changelog_parts = [
        beginning.strip(),
        '',
        heading_for_new_version,
        '',
        release_contents,
        '',
        rest
    ]

    with open(CHANGELOG_FILE, 'w') as o:
        o.write('\n'.join(new_changelog_parts))

    # Update the version specified in build.sbt.  We're looking to replace
    # a line of the form:
    #
    #       version := "x.y.z"
    #
    lines = list(open(BUILD_SBT))
    for idx, l in enumerate(lines):
        if l.startswith('version := '):
            lines[idx] = 'version := "%s"\n' % new_version_string.strip('v')
            break
    else:  # no break
        raise RuntimeError('Never updated version in build.sbt?')

    with open(BUILD_SBT, 'w') as f:
        f.write(''.join(lines))

    return release_type


def update_for_pending_scala_release():
    release_type = update_changelog_and_scala_version()

    git('rm', RELEASE_FILE)
    git('add', CHANGELOG_FILE)
    git('add', BUILD_SBT)

    git(
        'commit',
        '-m', 'Bump version to %s and update changelog\n\n[skip ci]' % (
            new_version(release_type))
    )
    git('tag', new_version(release_type))


def prepare_to_do_scala_release():
    last_release = latest_version()

    print('Latest released version: %s' % last_release)

    HEAD = hash_for_name('HEAD')
    MASTER = hash_for_name('origin/master')

    print('Current head:   %s' % HEAD)
    print('Current master: %s' % MASTER)

    on_master = is_ancestor(HEAD, MASTER)

    if not on_master:
        print('Trying to release while not on master?')
        sys.exit(1)

    if has_release():
        print('Updating changelog and version')
        update_for_pending_scala_release()
    else:
        print('Not releasing due to no release file')
        sys.exit(0)
